#include "aim.h"
#include "arithm.h"

struct AIMData_t aimData;
struct AIMControl_t aimControl;
struct AIMResult_t aimResult;

/*
* saves to aimRawBuffer[]
* TOP or BOTTOM entry (possibly both) according to disp_state and configured channel
* No filtering. Does not calculate finalResult (see getAimDisplayValue).
* But calculates Peak value of top/0 and bottom/0 (reset at reboot or powerup).
*/
void aim_save(void)
{
	uint16_t value, *p;
	uint8_t channel, confidx;

	aimResult.aimCount++;
	value = (aimData.value_h << 8) | aimData.value_l;
	/* top: */
	confidx = config.disp_state & 3; /* 0-3 */
	channel = aimData.channel;

	if( channel == config.aim_channel[ confidx ] ){
		aimResult.aimRawBuffer[ AIM_DISPLAY_TOP ] = value;
	}

	p = &(aimResult.aimRawBuffer[ AIM_DISPLAY_PEAKTOP ]);
	/* check if AIM_DISPLAY_PEAKTOP needs to be updated */
	if( channel == config.aim_channel[ 0 ] && value > *p ){
		*p = value;
	}

	confidx = 4;
	/* p = & aimResult.aimRawBuffer[ AIM_DISPLAY_PEAKBOTTOM ] ; replaced by ++p below */
	/* check if AIM_DISPLAY_PEAKBOTTOM needs to be updated */
	if ( channel == config.aim_channel[ confidx ] && value > *(++p) ){
		*p = value;
	}
	/* bottom: */
	if(config.disp_state < 4){
		confidx += (config.disp_state & 3); /* 4-7 */
	}
	if( channel == config.aim_channel[ confidx ] ){
		aimResult.aimRawBuffer[ AIM_DISPLAY_BOTTOM ] = value;
	}
	
}

uint8_t aim_feed( uint8_t c )
{
    if( (c == AIM_CONST) && (aimControl.state == AIM_UNKNOWN) ){
	uint8_t prev = aimControl.prevByte;
        aimData.channel = prev;
        aimData.sum = prev + c;
        aimControl.state = AIM_START;
        return aimControl.state;
    }
    aimControl.prevByte = c;

    switch( aimControl.state )
    {
        case AIM_START:
        {
            aimData.value_h = c;
            aimData.sum += c;
            aimControl.state = AIM_DATA_LOW;
            return aimControl.state;
        }
        case AIM_DATA_LOW:
        {
            aimData.value_l = c;
            aimData.sum += c;
            aimControl.state = AIM_SUM;
            return aimControl.state;
        }
        case AIM_SUM:
            aimControl.state = AIM_UNKNOWN;
            if( aimData.sum == c ){
		aim_save();
                return AIM_OK;
            }else{
                /* printf("CHK error: %02X,%02X\n", aimData.sum, c); */
                return AIM_ERROR;
            }
    }

    return aimControl.state;
}

/* should be asm */
uint16_t swapbyte( uint16_t in )
{
	uint8_t up, low;
	up = in & 0xFF;
	low = in >> 8;
	return (up << 8) | low;
}
/*
* pos = 0..3 top
* pos = 4..7 bottom
* calculates aimFinalResult: Ax>>7+B-8192
*/
int16_t getAimDisplayValue(uint8_t pos)
{
	uint32_t value32;
	uint16_t rawValue, b;
	int16_t value16;
	uint8_t pos01, a;
	pos &= 7; /* might not be necessary of &=15 might be more appropriate */
	pos01 = (pos & 4) ? 1 : 0; /* 0=>TOP, 1=>BOTTOM */
	if ( config.disp_state == 4 ){ /* for displaystate 4 */
		pos01 += 2; /* use aimRawBuffer PEAK values instead */
		pos = pos & 4; /* use the proper TOP/0 or BOTTOM/0 constants */
	}
	rawValue = aimResult.aimRawBuffer[pos01];
	a = config.aim_a[pos];
	value32 = mult16_8( rawValue, a );
	value32 >>= 7;
	value16 = value32; /* we could check overflow */
	b = swapbyte( config.aim_b[pos] );
	/* b &= 0x3FFF; bit15:14 reserved for DOT position */
	value16 += b;
	value16 -= 8192;

	/* aimResult.aimFinalResult[pos01] = value16; just for debugging */
	return value16;
}

/* Unpacks selected 2 bits of packed (originally from config.aim_dot[])
* idx (lower 2 bits) => selection of bits from packed
* 0 => 1:0
* 1 => 3:2
* 2 => 5:4
* 3 => 7:6
* Than the 2 bit values mean:
* 0: 1 digit after dot
* 1: 2 digits after dot
* 2: 3 digits after dot
* 3: NO dot
* Could be implemented in assembly.
*/
uint8_t aim_unpack( uint8_t idx, uint8_t packed )
{
	if( idx & 2 ){
		packed >>= 4; /* swap */
	}
	if( idx & 1 ){
		packed >>=2;
	}
	packed &= 3;
	/* 3 means no dot: */
	return (packed == 3 ? 0xFE : packed);
}
