/* VEMS ROUND Copyright 2005 Marcell Gal (XDSL Ltd Hungary) */

#ifndef __ROUND_C__
#define __ROUND_C__

#define MAINOBJ
#include "round.h"
#include "comm.h"
#include "adc.h"
#include "wbo2.h"
#include "ringbuf.h"
#include "tlc5920.h"
#include "menu.h"
#include "global.h"
#include "ioinit.c"
#include "softpwm.h"
#include "egt.h"
#include "storage.h"
#include "comparator.h"
#include "aim.h"

#include <avr/io.h>
#include <inttypes.h>
#include <avr/wdt.h>
#include <stdlib.h>
/* #include <avr/signal.h> */
#include <avr/interrupt.h>
#include <avr/sfr_defs.h>
#include <avr/pgmspace.h>

/* uint8_t tlc_counter; */

/* weight (number of high bits) of numbers 0..15 */
uint8_t w16[] PROGMEM = { 0,1,1,2,  1,2,2,3, 1,2,2,3, 2,3,3,4 };

/* should be in eventqueue.c when we add it back */
struct etime_t etime;
struct debug_t debug;

struct softpwm_t boost_pwm;
struct softpwm_t vout_pwm;

struct elapsed8_t elapsed_display;
struct config_t config;
struct runtime_t runtime;

uint8_t pwm_counter;
uint8_t pwm_counter_mem;

/* BOOST is PORTC3 (can also be used as ADC input) */
#define BOOST_PORT PORTC
#define BOOST_DDR DDRC
#define BOOST_MASK 8
/* see runtime.status if changing BOOST_MASK */

#define OUTPUTFREEZE_CYCLES 255
/** sets boost output (activates FET) according to
* config.out_minrpm ... variables
* outputfreeze = 255 after any change to prevent prelling
* TODO: BOOST_DDR and watch LSU4_PRESENT interference
*/
void boost_output(void)
{
	uint16_t t16;
	uint8_t output;
	uint8_t rs, rpm, egt, ain;
	t16 = runtime.rpm_vhr;
	rpm = t16 >= 16320 ? 0xFF : t16 >> 6; /* change to 64 RPM resolution */

	t16 = runtime.current_egt;
	egt = t16 >= 2040 ? 0xFF : t16 >> 3; /* 8C resolution */

	t16 = runtime.analin;
	ain = t16 >= 2040 ? 0xFF : t16 >> 3; /* 8 whatever resolution */

	if( rpm >= config.out_minrpm && rpm <= config.out_maxrpm
		&& egt >= config.out_minegt && egt <= config.out_maxegt
		&& ain >= config.out_minain && ain <= config.out_maxain ){
		output =1;
	} else {
		output = 0;
	}

	/* set runtime.status and output */
	rs = runtime.status;
	if( output ){ /* now ON */
		if( (rs & BOOST_MASK) == 0 ){ /* earlier OFF => changed */
			rs |= BOOST_MASK;
			BOOST_PORT |= BOOST_MASK;
			runtime.outputfreeze = OUTPUTFREEZE_CYCLES;
		}
	} else { /* now OFF */
		if( rs & BOOST_MASK){ /* earlier ON => changed */
			rs &= ~BOOST_MASK;
			BOOST_PORT &= ~BOOST_MASK;
			runtime.outputfreeze = OUTPUTFREEZE_CYCLES;
		}
	}
	runtime.status = rs;
} /* boost_output() */

/***************************************************************************/
/*** main entry point                                                    ***/
/***************************************************************************/
/*!
 * \ingroup g_main_loop
 * \brief Main entry point and cycle loop for fuel processing.
 *
 * \return none
 */
int main(void) {

	/*
  	 * init HW and misc. variables
	 * we keep interrupts disabled until everything is initialized
	 */
	init();

	/* enable interrupts */
	sei();

	/* main loop */
	for (;;) {
		uint8_t t8 = 0;

		/*
		 * update AFR display
		 * update EGT display
		 * rs232 buffer is always processed
		 */
		ringbuf_shift_proc(&menu_rs232, &comm_recbuf);

		/*
		 * collect data for the transmit buffer
		 * a lightweight scheduler could be included
		 */

		/* TODO: why is this commented ? */
		ringbuf_fill(&comm_sndbuf);

		/*
		tlc_counter++;
		if(tlc_enabled()){ */ /* not needed, set refresh_period=0xFF */
			/* if(tlc_counter > tlc5920.refresh_period){
				tlc5920.refresh_period = 0;
				*/
				/* TODO: dimming */
				/*
				tlc5920_action();
			}
		}
		*/

#ifdef WBO2
		/* nernst evaluation, pump and heater PID and actuation: */
		wbo2_heater();
		/*
		 * calculate the lambda / o2 results from time to time
		 * (appr. 50..100 msec)
		 */
#endif

	analin_calculate();
	if( runtime.outputfreeze ){ /* do not touch output, count down */
		runtime.outputfreeze --;
	} else { /* set output */
		/* boost_output(); TODO: enable this. Only if boostpwm inactive ?*/
	}

	/*
	 * If the button is just pressed the user expects imediate action.
	 * otherwise we don't like the displayed values flipping at 40 Hz
	 */
	if( (check_button( &button ) & ( _BV(BUTTON_CHANGED) | _BV( BUTTON_PRESSED )))
		== ( _BV(BUTTON_CHANGED) | _BV( BUTTON_PRESSED ) )
		){
		button.state &= ~_BV(BUTTON_CHANGED); /* reset the flag */
		t8 = 0;
	} else {
		t8 = 255;	/* display every 300..400 msec would be better */
	}

	if(softelapsed8( &elapsed_display, t8 )){ /* display handling */

        if(config.segment3 == 0xA5) { /* Bottom AIM channel */
            uint8_t aim_config_idx = 4 | ((config.disp_state >>2) & 3);
            uint16_t val = getAimDisplayValue(aim_config_idx);
            int2BCDn1(val, 3, DIGIT4, 0xFE);
            /* aim_active = 1; AIMFLAG=Y; */
        }else{ /* Normal mode */
            update_temps();
            display_bottom();
        }

        if(config.segment0 == 0xA5) { /* Top AIM channel */
            uint8_t aim_config_idx = config.disp_state & 3;
            uint16_t val = getAimDisplayValue(aim_config_idx);
            int2BCDn1(val, 0, DIGIT3, 0xFE);
            /* AIMFLAG=Y; */
        }else{ /* Normal mode */
#ifdef WBO2
            /* on the 3+4 digit display, the 3 is more suited for WBO2 */
            #define WBO2_DISPLAY_INDEX 0
            /* note that wbo2 display is ideally independent of the control! */
            wbo2_display( WBO2_DISPLAY_INDEX, &wbo2 );
#endif
		}
	} /* end of display handling */

		if(boostpwm_enabled()) {
			/*
			 * BOOSTOUT current limitation
			 * enabling this WILL disturb LSU4_PRESENT measurement
			 */
			if(softpwm_act(&boost_pwm)){
				BOOST_DDR |= BOOST_MASK;	/* set port to output */
				BOOST_PORT |= BOOST_MASK;	/* output high */
			} else {
				/* use similar strategy (regarding lsu_pres ADC) as for the WBO2 heater */
				BOOST_PORT &= ~BOOST_MASK;	/* output low */
				/*
				 * some delay (1usec ?) to discharge the FET (through 510 Ohm) is desirable
				 * before changing to input (that is 10k external pulldown)
				 */
				BOOST_DDR &= ~BOOST_MASK; 	/* switch to input again */
			}
		}

#define VOUT_PORT PORTB
#define VOUT_DDR DDRB
#define VOUT_MASK 2

		if(softpwm_act(&vout_pwm)){
				VOUT_PORT |= VOUT_MASK;		/* output high */
		} else {
				VOUT_PORT &= ~VOUT_MASK;	/* output low */
				/* VOUT_DDR &= ~VOUT_MASK; */	/* switch to input again */
		}

		/* make sure time is updated (upper 16 bit also), even without trigger */
		(void) getTime();

		t8 = comparator.now_upper;
		t8 -= (comparator.captured32_prev >> 16);
		if( t8 >= 4 ){
			/* no trigger for appr 4 * 262 msec : */
			runtime.rpm_vhr = 0;
		} else {
			rpm_calc();
		}

  		store_control();
	} /* main loop */
	/*@notreached@*/
	return (0);
} /* main() */


/***************************************************************************/
/*** init                                                                ***/
/***************************************************************************/
/*!
 * \ingroup g_init
 * \brief Main Initialization for power-up.
 * All under cli();
 * \return none
 */
void init(void) {
	debug.disable_flags =			/* 0xFF | disables everything by default */
	_BV( DIS_FLAG_BOOSTPWM_DISABLED );	/* boost disabled to allow LSU4_PRES sampling */
	ioinit();
	/*
	 * menu system that is controllable from rs232 or ps2keyboard
	 * init que's before uart
	 */
	ringbuf_init(&comm_recbuf);
	ringbuf_init(&comm_sndbuf);
	menu_init(&menu_rs232, MENU_ZERO );	/* MENU_mmm */

	/* init rs232 */
	initUART();
	tlc5920_init();

	(void)pushfuncpar(&srl_generic_P, 8, (uint8_t *)PSTR("Booted\r\n"));

	init_storage();
  	load_variables();			/* load default config */
#ifdef WBO2
	init_wbo2();
#endif
	/* egt_init(); empty anyway, for now */

	softpwm_init(&boost_pwm, 0);
	softpwm_init(&vout_pwm, 0);

	/* beware: timing_init() must be called after init_wbo2() */
	/* timing_init(); */
	/* should we call this earlier?  We want to catch early triggers but
	 * we really would not like to use unitialized variables. Cannot do much, in cli() anyway...
	 */

	/* initialize rtc variables: */
	/* No RTC hardward on Round
	rtc.tsec = 10;
	rtc.adc_clk = 1;
	*/

	init_tcnt0();
	init_tcnt2();
	/* mainly for eventqueue: (initialize after the others) */
	init_tcnt1();
	counter_sync();

	/* init the sensor array */
	init_sensors();

	init_and_start_adc();

	/* init ain */
	init_ain();
} /* init() */

#endif /* __ROUND_C__ */


